@@ -2,9 +2,13 @@
 //!
 //! Implements Bayesian program learning and genetic programming.
 
+extern crate crossbeam_channel;
 extern crate itertools;
 #[macro_use]
+extern crate lazy_static;
+#[macro_use]
 extern crate nom;
+extern crate num_cpus;
 #[macro_use]
 extern crate polytype;
 extern crate rand;
